# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import os
import re
import signal
import subprocess
import threading
import time
import typing
import unittest

from cotyledon import oslo_config_glue


if os.name == "posix":

    def pid_exists(pid: int) -> bool:
        """Check whether pid exists in the current process table."""
        import errno  # noqa: PLC0415

        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            return e.errno == errno.EPERM
        else:
            return True
else:

    def pid_exists(pid: int) -> bool:
        import ctypes  # noqa: PLC0415

        kernel32 = ctypes.windll.kernel32  # type: ignore[attr-defined]
        synchronize = 0x100000

        process = kernel32.OpenProcess(synchronize, 0, pid)
        if process != 0:
            kernel32.CloseHandle(process)
            return True
        return False


class Base(unittest.TestCase):
    name: typing.ClassVar[str]

    def setUp(self) -> None:
        super().setUp()

        self.lines: list[bytes] = []

        examplepy = os.path.join(os.path.dirname(__file__), "examples.py")
        self.subp = subprocess.Popen(
            ["python", examplepy, self.name],
            stdout=subprocess.PIPE,
            creationflags=0
            if os.name == "posix"
            else subprocess.CREATE_NEW_PROCESS_GROUP,  # type: ignore[attr-defined]
            preexec_fn=os.setsid if os.name == "posix" else None,  # noqa:  PLW1509
        )

        self.t = threading.Thread(target=self.readlog)
        self.t.daemon = True
        self.t.start()

    def readlog(self) -> None:
        if self.subp.stdout is None:
            raise RuntimeError("Unexpected subprocess setup")  # noqa: EM101,TRY003
        while True:
            try:
                line = self.subp.stdout.readline().strip()
            except OSError:
                return
            if not line:
                continue
            # Keep only logger message
            if b":cotyledon." not in line:
                print(b"?? " + line)  # noqa: T201
                continue

            print(b"<> " + line)  # noqa: T201
            self.lines.append(line)

    def tearDown(self) -> None:
        if self.subp.poll() is None:
            self.subp.kill()
        super().tearDown()

    def get_lines(self, number: int | None = None) -> list[bytes]:
        if number is not None:
            while len(self.lines) < number:
                time.sleep(0.1)
            lines = self.lines[:number]
            del self.lines[:number]
            return lines
        self.subp.wait()
        # Wait children to terminate
        return self.lines

    @staticmethod
    def hide_pids(lines: list[bytes]) -> list[bytes]:
        return [
            re.sub(
                rb"Child \d+",
                b"Child XXXX",
                re.sub(rb" \[[^\]]*\]", b" [XXXX]", line),
            )
            for line in lines
        ]

    @staticmethod
    def get_pid(line: bytes) -> int:
        try:
            return int(line.split()[-1][1:-1])
        except Exception as exc:
            msg = f"Fail to find pid in {line.split()}"
            raise RuntimeError(msg) from exc


class TestCotyledon(Base):
    name = "example_app"

    def assert_everything_has_started(self) -> None:
        lines = sorted(self.get_lines(7))
        self.pid_heavy_1 = self.get_pid(lines[0])
        self.pid_heavy_2 = self.get_pid(lines[1])
        self.pid_light_1 = self.get_pid(lines[2])
        lines = self.hide_pids(lines)
        assert lines == [
            b"DEBUG:cotyledon._service_worker:Run service heavy(0) [XXXX]",
            b"DEBUG:cotyledon._service_worker:Run service heavy(1) [XXXX]",
            b"DEBUG:cotyledon._service_worker:Run service light(0) [XXXX]",
            b"ERROR:cotyledon.tests.examples:heavy init",
            b"ERROR:cotyledon.tests.examples:heavy init",
            b"ERROR:cotyledon.tests.examples:heavy run",
            b"ERROR:cotyledon.tests.examples:heavy run",
        ]

        self.assert_everything_is_alive()

    def assert_everything_is_alive(self) -> None:
        assert pid_exists(self.subp.pid)
        assert pid_exists(self.pid_light_1)
        assert pid_exists(self.pid_heavy_1)
        assert pid_exists(self.pid_heavy_2)

    def assert_everything_is_dead(self, status: int = 0) -> None:
        assert status == self.subp.poll()
        assert not pid_exists(self.subp.pid)
        assert not pid_exists(self.pid_light_1)
        assert not pid_exists(self.pid_heavy_1)
        assert not pid_exists(self.pid_heavy_2)

    @unittest.skipIf(os.name == "posix", "no window support")
    def test_workflow_window(self) -> None:
        # NOTE(sileht): The window workflow is a bit different because
        # SIGTERM doesn't really exists and processes are killed with SIGINT
        # TODO(sileht): Implements SIGBREAK to have graceful exists

        self.assert_everything_has_started()
        # Ensure we restart with terminate method exit code
        os.kill(self.pid_heavy_1, signal.SIGTERM)
        lines = self.get_lines(4)
        lines = self.hide_pids(lines)
        assert lines == [
            b"INFO:cotyledon._service_manager:Child XXXX exited with status 15",
            b"ERROR:cotyledon.tests.examples:heavy init",
            b"DEBUG:cotyledon._service_worker:Run service heavy(0) [XXXX]",
            b"ERROR:cotyledon.tests.examples:heavy run",
        ]

        # Kill master process
        os.kill(self.subp.pid, signal.SIGTERM)
        time.sleep(1)
        lines = self.get_lines()
        lines = sorted(self.hide_pids(lines))
        assert lines == [
            b"ERROR:cotyledon.tests.examples:heavy terminate",
            b"ERROR:cotyledon.tests.examples:heavy terminate",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service heavy(0) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service heavy(1) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service light(0) [XXXX]",
            b"INFO:cotyledon._service_worker:Parent process has died unexpectedly, heavy(0) [XXXX] exiting",
            b"INFO:cotyledon._service_worker:Parent process has died unexpectedly, heavy(1) [XXXX] exiting",
            b"INFO:cotyledon._service_worker:Parent process has died unexpectedly, light(0) [XXXX] exiting",
        ]

        assert self.subp.poll() == 15

    @unittest.skipIf(os.name != "posix", "no posix support")
    def test_workflow(self) -> None:
        self.assert_everything_has_started()

        # Ensure we just call reload method
        os.kill(self.pid_heavy_1, signal.SIGHUP)
        assert self.get_lines(1) == [b"ERROR:cotyledon.tests.examples:heavy reload"]

        # Ensure we restart because reload method is missing
        os.kill(self.pid_light_1, signal.SIGHUP)
        lines = self.get_lines(3)
        self.pid_light_1 = self.get_pid(lines[-1])
        lines = self.hide_pids(lines)
        assert lines == [
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service light(0) [XXXX]",
            b"INFO:cotyledon._service_manager:Child XXXX exited with status 0",
            b"DEBUG:cotyledon._service_worker:Run service light(0) [XXXX]",
        ]

        # Ensure we restart with terminate method exit code
        os.kill(self.pid_heavy_1, signal.SIGTERM)
        lines = self.get_lines(6)
        self.pid_heavy_1 = self.get_pid(lines[-2])
        lines = self.hide_pids(lines)
        assert lines == [
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service heavy(0) [XXXX]",
            b"ERROR:cotyledon.tests.examples:heavy terminate",
            b"INFO:cotyledon._service_manager:Child XXXX exited with status 42",
            b"ERROR:cotyledon.tests.examples:heavy init",
            b"DEBUG:cotyledon._service_worker:Run service heavy(0) [XXXX]",
            b"ERROR:cotyledon.tests.examples:heavy run",
        ]

        # Ensure we restart when no terminate method
        os.kill(self.pid_light_1, signal.SIGTERM)
        lines = self.get_lines(3)
        self.pid_light_1 = self.get_pid(lines[-1])
        lines = self.hide_pids(lines)
        assert lines == [
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service light(0) [XXXX]",
            b"INFO:cotyledon._service_manager:Child XXXX exited with status 0",
            b"DEBUG:cotyledon._service_worker:Run service light(0) [XXXX]",
        ]

        # Ensure everything is still alive
        self.assert_everything_is_alive()

        # Kill master process
        os.kill(self.subp.pid, signal.SIGTERM)

        lines = self.get_lines()
        assert lines[-1] == b"DEBUG:cotyledon._service_manager:Shutdown finish"
        time.sleep(1)
        lines = sorted(self.hide_pids(lines))
        assert lines == [
            b"DEBUG:cotyledon._service_manager:Killing services with signal SIGTERM",
            b"DEBUG:cotyledon._service_manager:Shutdown finish",
            b"DEBUG:cotyledon._service_manager:Waiting services to terminate",
            b"ERROR:cotyledon.tests.examples:heavy terminate",
            b"ERROR:cotyledon.tests.examples:heavy terminate",
            b"ERROR:cotyledon.tests.examples:master terminate hook",
            b"ERROR:cotyledon.tests.examples:master terminate2 hook",
            b"INFO:cotyledon._service_manager:Caught SIGTERM signal, graceful exiting of master process",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service heavy(0) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service heavy(1) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service light(0) [XXXX]",
        ]

        self.assert_everything_is_dead()

    @unittest.skipIf(os.name != "posix", "http://bugs.python.org/issue18040")
    def test_sigint(self) -> None:
        self.assert_everything_has_started()
        os.kill(self.subp.pid, signal.SIGINT)
        time.sleep(1)
        lines = sorted(self.get_lines())
        lines = self.hide_pids(lines)
        assert lines == [
            b"INFO:cotyledon._service_manager:Caught SIGINT signal, instantaneous exiting",
            b"INFO:cotyledon._service_worker:Caught SIGINT signal, instantaneous exiting of service heavy(0) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGINT signal, instantaneous exiting of service heavy(1) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGINT signal, instantaneous exiting of service light(0) [XXXX]",
        ]
        self.assert_everything_is_dead(1)

    @unittest.skipIf(os.name != "posix", "no posix support")
    def test_sighup(self) -> None:
        self.assert_everything_has_started()
        os.kill(self.subp.pid, signal.SIGHUP)
        time.sleep(0.5)
        lines = sorted(self.get_lines(6))
        lines = self.hide_pids(lines)
        assert lines == [
            b"DEBUG:cotyledon._service_worker:Run service light(0) [XXXX]",
            b"ERROR:cotyledon.tests.examples:heavy reload",
            b"ERROR:cotyledon.tests.examples:heavy reload",
            b"ERROR:cotyledon.tests.examples:master reload hook",
            b"INFO:cotyledon._service_manager:Child XXXX exited with status 0",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service light(0) [XXXX]",
        ]

        os.kill(self.subp.pid, signal.SIGINT)
        time.sleep(0.5)
        self.assert_everything_is_dead(1)

    @unittest.skipIf(os.name != "posix", "no posix support")
    def test_sigkill(self) -> None:
        self.assert_everything_has_started()
        self.subp.kill()
        time.sleep(1)
        lines = sorted(self.get_lines())
        lines = self.hide_pids(lines)
        assert lines == [
            b"ERROR:cotyledon.tests.examples:heavy terminate",
            b"ERROR:cotyledon.tests.examples:heavy terminate",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service heavy(0) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service heavy(1) [XXXX]",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service light(0) [XXXX]",
            b"INFO:cotyledon._service_worker:Parent process has died unexpectedly, heavy(0) [XXXX] exiting",
            b"INFO:cotyledon._service_worker:Parent process has died unexpectedly, heavy(1) [XXXX] exiting",
            b"INFO:cotyledon._service_worker:Parent process has died unexpectedly, light(0) [XXXX] exiting",
        ]
        self.assert_everything_is_dead(-9)


class TestBadlyCodedCotyledon(Base):
    name = "badly_coded_app"

    @unittest.skipIf(os.name != "posix", "no posix support")
    def test_badly_coded(self) -> None:
        time.sleep(2)
        self.subp.terminate()
        time.sleep(2)
        assert self.subp.poll() == 0, self.get_lines()
        assert not pid_exists(self.subp.pid)


class TestBuggyCotyledon(Base):
    name = "buggy_app"

    @unittest.skipIf(os.name != "posix", "no posix support")
    def test_graceful_timeout_term(self) -> None:
        lines = self.get_lines(1)
        childpid = self.get_pid(lines[0])
        self.subp.terminate()
        time.sleep(2)
        assert self.subp.poll() == 0
        assert not pid_exists(self.subp.pid)
        assert not pid_exists(childpid)
        lines = self.hide_pids(self.get_lines())
        assert b"ERROR:cotyledon.tests.examples:time.sleep done" not in lines
        assert lines[-2:] == [
            b"INFO:cotyledon._service_worker:Graceful shutdown timeout (1) exceeded, exiting buggy(0) [XXXX] now.",
            b"DEBUG:cotyledon._service_manager:Shutdown finish",
        ]

    @unittest.skipIf(os.name != "posix", "no posix support")
    def test_graceful_timeout_kill(self) -> None:
        lines = self.get_lines(1)
        childpid = self.get_pid(lines[0])
        self.subp.kill()
        time.sleep(2)
        assert self.subp.poll() == -9
        assert not pid_exists(self.subp.pid)
        assert not pid_exists(childpid)
        lines = self.hide_pids(self.get_lines())
        assert b"ERROR:cotyledon.tests.examples:time.sleep done" not in lines
        assert lines[-3:] == [
            b"INFO:cotyledon._service_worker:Parent process has died unexpectedly, buggy(0) [XXXX] exiting",
            b"INFO:cotyledon._service_worker:Caught SIGTERM signal, graceful exiting of service buggy(0) [XXXX]",
            b"INFO:cotyledon._service_worker:Graceful shutdown timeout (1) exceeded, exiting buggy(0) [XXXX] now.",
        ]


class TestOsloCotyledon(Base):
    name = "oslo_app"

    def test_options(self) -> None:
        options = oslo_config_glue.list_opts()
        assert len(options) == 1
        assert None is options[0][0]
        assert len(options[0][1]) == 2

        lines = self.get_lines(1)
        assert b"DEBUG:cotyledon.oslo_config_glue:Full set of CONF:" in lines
        self.subp.terminate()


class TestTermDuringStartupCotyledon(Base):
    name = "sigterm_during_init"

    def test_sigterm(self) -> None:
        lines = self.hide_pids(self.get_lines())
        assert b"DEBUG:cotyledon._service_manager:Shutdown finish" in lines



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import multiprocessing
import typing
from unittest import TestCase
from unittest import mock

import pytest

import cotyledon
from cotyledon import _utils


P = typing.ParamSpec("P")
R = typing.TypeVar("R")


class FakeService(cotyledon.Service):
    pass


class SomeTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        cotyledon.ServiceManager._process_runner_already_created = False

    def test_forking_slowdown(self) -> None:
        sm = cotyledon.ServiceManager()
        sm.add(FakeService, workers=3)
        with mock.patch("time.sleep") as sleep:
            sm._slowdown_respawn_if_needed()
            sm._slowdown_respawn_if_needed()
            sm._slowdown_respawn_if_needed()
            # We simulatge 3 more spawn
            sm._slowdown_respawn_if_needed()
            sm._slowdown_respawn_if_needed()
            sm._slowdown_respawn_if_needed()
            assert len(sleep.mock_calls) == 6

    def test_invalid_service(self) -> None:
        sm = cotyledon.ServiceManager()

        self.assert_raises_msg(
            TypeError,
            "'service' must be a callable",
            sm.add,
            "foo",  # type: ignore[arg-type]
        )
        self.assert_raises_msg(
            ValueError,
            "'workers' must be an int >= 1, not: None (NoneType)",
            sm.add,
            FakeService,
            workers=None,  # type: ignore[arg-type]
        )
        self.assert_raises_msg(
            ValueError,
            "'workers' must be an int >= 1, not: -2 (int)",
            sm.add,
            FakeService,
            workers=-2,
        )

        oid = sm.add(FakeService, workers=3)
        self.assert_raises_msg(
            ValueError,
            "'workers' must be an int >= -2, not: -5 (int)",
            sm.reconfigure,
            oid,
            workers=-5,
        )
        self.assert_raises_msg(
            ValueError,
            "notexists service id doesn't exists",
            sm.reconfigure,
            "notexists",  # type: ignore[arg-type]
            workers=-1,
        )

    @staticmethod
    def assert_raises_msg(
        exc: type[Exception] | tuple[type[Exception], ...],
        msg: str,
        func: typing.Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        with pytest.raises(exc) as exc_info:
            func(*args, **kwargs)
        assert msg == str(exc_info.value)


class TestMultiprocessing(TestCase):
    def setUp(self) -> None:
        super().setUp()
        cotyledon.ServiceManager._process_runner_already_created = False

    def test_multiprocessing_context_default(self) -> None:
        """Test that ServiceManager uses multiprocessing (fork) by default"""
        sm = cotyledon.ServiceManager()
        # Default should be multiprocessing context (fork-compatible)
        assert sm.mp_context == multiprocessing.get_context()

    def test_multiprocessing_context_constructor(self) -> None:
        """Test that mp_context parameter in constructor works"""
        spawn_ctx = multiprocessing.get_context("spawn")
        sm = cotyledon.ServiceManager(mp_context=spawn_ctx)
        assert sm.mp_context is spawn_ctx

    def test_spawn_process_with_context(self) -> None:
        """Test that spawn_process uses the provided context"""

        def dummy_target() -> None:
            pass

        # Test with default context (None)
        mock_proc_instance = mock.Mock()
        mock_proc_instance.start = mock.Mock()
        with mock.patch(
            "multiprocessing.Process",
            return_value=mock_proc_instance,
        ) as mock_process:
            _utils.spawn_process(dummy_target)
            # Should use multiprocessing.Process
            mock_process.assert_called_once()
            mock_proc_instance.start.assert_called_once()

        # Test with explicit context
        spawn_ctx = multiprocessing.get_context("spawn")
        mock_ctx_proc_instance = mock.Mock()
        mock_ctx_proc_instance.start = mock.Mock()
        with mock.patch.object(
            spawn_ctx,
            "Process",
            return_value=mock_ctx_proc_instance,
        ) as mock_ctx_process:
            _utils.spawn_process(dummy_target, ctx=spawn_ctx)
            # Should use ctx.Process
            mock_ctx_process.assert_called_once()
            mock_ctx_proc_instance.start.assert_called_once()

    def test_spawn_process_backward_compatible(self) -> None:
        """Test that spawn_process without ctx is backward compatible"""

        def dummy_target() -> None:
            pass

        # When ctx is None, should default to multiprocessing
        mock_proc_instance = mock.Mock()
        mock_proc_instance.start = mock.Mock()
        with mock.patch(
            "multiprocessing.Process",
            return_value=mock_proc_instance,
        ) as mock_process:
            _utils.spawn_process(dummy_target, ctx=None)
            mock_process.assert_called_once()
            mock_proc_instance.start.assert_called_once()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
import os
import signal
import typing


if typing.TYPE_CHECKING:
    from cotyledon import types


LOG = logging.getLogger(__name__)


class Service:
    """Base class for a service

    This class will be executed in a new child process/worker
    :py:class:`ServiceWorker` of a :py:class:`ServiceManager`. It registers
    signals to manager the reloading and the ending of the process.

    Methods :py:meth:`run`, :py:meth:`terminate` and :py:meth:`reload` are
    optional.
    """

    name: typing.ClassVar[str | None] = None
    """Service name used in the process title and the log messages in
    additional of the worker_id."""

    graceful_shutdown_timeout: typing.ClassVar[int | None] = None
    """Timeout after which a gracefully shutdown service will exit. zero means
    endless wait. None means same as ServiceManager that launch the service"""

    def __init_subclass__(cls, **kwargs: typing.Any) -> None:  # noqa: ANN401
        super().__init_subclass__(**kwargs)
        if cls.name is None:
            cls.name = cls.__name__

    def __init__(self, worker_id: types.WorkerId) -> None:
        """Create a new Service

        :param worker_id: the identifier of this service instance
        :type worker_id: int

        The identifier of the worker can be used for workload repartition
        because it's consistent and always the same.

        For example, if the number of workers for this service is 3,
        one will got 0, the second got 1 and the last got 2.
        if worker_id 1 died, the new spawned process will got 1 again.
        """
        super().__init__()
        self.worker_id = worker_id

    def terminate(self) -> None:
        """Gracefully shutdown the service

        This method will be executed when the Service has to shutdown cleanly.

        If not implemented the process will just end with status 0.

        To customize the exit code, the :py:class:`SystemExit` exception can be
        used.

        Any exceptions raised by this method will be logged and the worker will
        exit with status 1.
        """

    def reload(self) -> None:  # noqa: PLR6301
        """Reloading of the service

        This method will be executed when the Service receives a SIGHUP.

        If not implemented the process will just end with status 0 and
        :py:class:`ServiceRunner` will start a new fresh process for this
        service with the same worker_id.

        Any exceptions raised by this method will be logged and the worker will
        exit with status 1.
        """
        os.kill(os.getpid(), signal.SIGTERM)

    def run(self) -> None:
        """Method representing the service activity

        If not implemented the process will just wait to receive an ending
        signal.

        This method is ran into the thread and can block or return as needed

        Any exceptions raised by this method will be logged and the worker will
        exit with status 1.
        """



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import logging
import os
import random
import signal
import sys
import threading
import typing

from cotyledon import _service
from cotyledon import _utils
from cotyledon import types as t


if typing.TYPE_CHECKING:
    import multiprocessing.connection

LOG = logging.getLogger(__name__)

MultiprocessingPipe: typing.TypeAlias = tuple[
    "multiprocessing.connection.Connection",
    "multiprocessing.connection.Connection",
]


OnNewWorkerHook: typing.TypeAlias = typing.Callable[
    [t.ServiceId, t.WorkerId, _service.Service],
    None,
]

P = typing.ParamSpec("P")
R = typing.TypeVar("R")

ServiceT = typing.TypeVar("ServiceT", bound=_service.Service)


class ServiceConfig(typing.Generic[ServiceT, P]):
    def __init__(
        self,
        service_id: t.ServiceId,
        service: type[ServiceT],
        workers: int,
        args: typing.Any,  # noqa: ANN401
        kwargs: typing.Any,  # noqa: ANN401
    ) -> None:
        self.service = service
        self.workers = workers
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs
        self.service_id = service_id

    def build(self, worker_id: t.WorkerId) -> ServiceT:
        return self.service(worker_id, *self.args, **self.kwargs)


class ServiceWorker(_utils.SignalManager):
    """Service Worker Wrapper

    This represents the child process spawned by ServiceManager

    All methods implemented here, must run in the main threads
    """

    @classmethod
    def create_and_wait(  # noqa: PLR0913,PLR0917
        cls,
        started_event: multiprocessing.synchronize.Event,
        config: ServiceConfig[typing.Any, typing.Any],
        service_id: t.ServiceId,
        worker_id: t.WorkerId,
        parent_pipe: MultiprocessingPipe,
        started_hooks: list[OnNewWorkerHook],
        graceful_shutdown_timeout: int,
    ) -> None:
        sw = cls(
            config,
            service_id,
            worker_id,
            parent_pipe,
            started_hooks,
            graceful_shutdown_timeout,
        )
        started_event.set()
        sw.wait_forever()

    def __init__(  # noqa: PLR0917, PLR0913
        self,
        config: ServiceConfig[typing.Any, typing.Any],
        service_id: t.ServiceId,
        worker_id: t.WorkerId,
        parent_pipe: MultiprocessingPipe,
        started_hooks: list[OnNewWorkerHook],
        graceful_shutdown_timeout: int,
    ) -> None:
        super().__init__()
        self._ready = threading.Event()
        self._signal_lock = threading.Lock()

        _utils.spawn(self._watch_parent_process, parent_pipe)

        # Reseed random number generator
        random.seed()

        self.service: _service.Service = config.build(worker_id)
        # Backward compat: Just in case the __init__ is not called on the service
        self.service.worker_id = worker_id

        self.title = f"{self.service.name}({worker_id}) [{os.getpid()}]"
        self._graceful_shutdown_timeout = graceful_shutdown_timeout

        # Set process title
        _utils.set_process_title(
            f"{_utils.get_process_name()}: {self.service.name} worker({worker_id})",
        )

        # We are ready tell them
        self._ready.set()
        _utils.run_hooks(
            "new_worker",
            started_hooks,
            service_id,
            worker_id,
            self.service,
        )

    def _get_graceful_shutdown_timeout(self) -> int:
        if self.service.graceful_shutdown_timeout is None:
            return self._graceful_shutdown_timeout

        return self.service.graceful_shutdown_timeout

    def _watch_parent_process(self, parent_pipe: MultiprocessingPipe) -> None:
        # This will block until the write end is closed when the parent
        # dies unexpectedly
        parent_pipe[1].close()
        with contextlib.suppress(EOFError):
            parent_pipe[0].recv()

        if self._ready.is_set():
            LOG.info("Parent process has died unexpectedly, %s exiting", self.title)
            if os.name == "posix":
                os.kill(os.getpid(), signal.SIGTERM)
            else:
                # Fallback to process signal later
                self._on_signal_received(signal.SIGTERM)
        else:
            os._exit(0)

    def _alarm(self) -> None:
        LOG.info(
            "Graceful shutdown timeout (%d) exceeded, exiting %s now.",
            self._get_graceful_shutdown_timeout(),
            self.title,
        )
        os._exit(1)

    def _fast_exit(self) -> None:
        LOG.info(
            "Caught SIGINT signal, instantaneous exiting of service %s",
            self.title,
        )
        os._exit(1)

    def _on_signal_received(self, sig: int) -> None:
        # Code below must not block to return to select.select() and catch
        # next signals
        if sig == _utils.SIGALRM:
            self._alarm()
        if sig == signal.SIGINT:
            self._fast_exit()
        elif sig == signal.SIGTERM:
            LOG.info(
                "Caught SIGTERM signal, graceful exiting of service %s",
                self.title,
            )

            timeout = self._get_graceful_shutdown_timeout()
            if timeout > 0:
                if os.name == "posix":
                    signal.alarm(timeout)
                else:
                    threading.Timer(timeout, self._alarm).start()
            _utils.spawn(self._service_terminate)
        elif sig == _utils.SIGHUP:
            _utils.spawn(self._service_reload)

    def _service_run(self) -> None:
        with _utils.exit_on_exception():
            self.service.run()

    def _service_terminate(self) -> None:
        with _utils.exit_on_exception(), self._signal_lock:
            self.service.terminate()
            sys.exit(0)

    def _service_reload(self) -> None:
        with _utils.exit_on_exception():
            if self._signal_lock.acquire(blocking=False):
                try:
                    self.service.reload()
                finally:
                    self._signal_lock.release()

    def wait_forever(self) -> None:
        LOG.debug("Run service %s", self.title)
        _utils.spawn(self._service_run)
        super()._wait_forever()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import errno
import logging
import multiprocessing
import os
import select
import signal
import sys
import threading
import typing


if typing.TYPE_CHECKING:
    import types


if os.name == "posix":
    import fcntl

LOG = logging.getLogger(__name__)


_SIGNAL_TO_NAME: dict[int, str] = {
    getattr(signal, name): name
    for name in dir(signal)
    if name.startswith("SIG") and isinstance(getattr(signal, name), signal.Signals)
}

SIGNAL_WAKEUP_FD_READ_SIZE = 4096


def signal_to_name(sig: int) -> str:
    return _SIGNAL_TO_NAME.get(sig, str(sig))


P = typing.ParamSpec("P")
R = typing.TypeVar("R")


def spawn(
    target: typing.Callable[P, R],
    *args: P.args,
    **kwargs: P.kwargs,
) -> threading.Thread:
    t = threading.Thread(target=target, args=args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t


def check_workers(workers: int, minimum: int) -> None:
    if not isinstance(workers, int) or workers < minimum:
        msg = f"'workers' must be an int >= {minimum}, not: {workers} ({type(workers).__name__})"
        raise ValueError(msg)


def check_callable(
    thing: typing.Any,  # noqa: ANN401
    name: str,
) -> None:
    if not callable(thing):
        msg = f"'{name}' must be a callable"
        raise TypeError(msg)


def spawn_process(
    target: typing.Callable[P, R],
    ctx: multiprocessing.context.BaseContext | None = None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> multiprocessing.Process:
    # typeshed doesn't expose "Process" on BaseContext, but it's there at runtime
    process_factory = (
        multiprocessing.Process if ctx is None else ctx.Process  # type: ignore[attr-defined]
    )
    p = process_factory(
        target=target,
        args=args,
        kwargs=kwargs,
    )
    p.start()
    return p


_setproctitle: typing.Callable[[str], None] | None
try:
    from setproctitle import setproctitle as _setproctitle
except ImportError:
    _setproctitle = None


def set_process_title(title: str) -> None:
    if _setproctitle is None:
        return

    _setproctitle(title)


def get_process_name() -> str:
    return os.path.basename(sys.argv[0])


def run_hooks(
    name: str,
    hooks: list[typing.Callable[P, R]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> None:
    try:
        for hook in hooks:
            hook(*args, **kwargs)
    except Exception:
        LOG.exception("Exception raised during %s hooks", name)


@contextlib.contextmanager
def exit_on_exception() -> typing.Iterator[None]:
    try:
        yield
    except SystemExit as exc:
        # mypy is sick, it's a int, not int|str|None
        os._exit(exc.code)  # type: ignore[arg-type]
    except BaseException:
        LOG.exception("Unhandled exception")
        os._exit(2)


if os.name == "posix":
    SIGALRM = signal.SIGALRM
    SIGHUP = signal.SIGHUP
    SIGCHLD = signal.SIGCHLD
else:
    SIGALRM = SIGHUP = SIGCHLD = None  # type: ignore[assignment]


class SignalManager:
    def __init__(self) -> None:
        # Setup signal fd, this allows signal to behave correctly
        if os.name == "posix":
            self.signal_pipe_r, self.signal_pipe_w = os.pipe()
            self._set_nonblock(self.signal_pipe_r)
            self._set_nonblock(self.signal_pipe_w)
            self._set_autoclose(self.signal_pipe_r)
            self._set_autoclose(self.signal_pipe_w)
            signal.set_wakeup_fd(self.signal_pipe_w)

        if os.name == "posix":
            signal.signal(signal.SIGCHLD, signal.SIG_DFL)
            signal.signal(signal.SIGINT, self._signal_catcher)
            signal.signal(signal.SIGTERM, self._signal_catcher)
            signal.signal(signal.SIGALRM, self._signal_catcher)
            signal.signal(signal.SIGHUP, self._signal_catcher)
        else:
            signal.signal(signal.SIGINT, self._signal_catcher)
            # currently a noop on window...
            signal.signal(signal.SIGTERM, self._signal_catcher)
            # TODO(sileht): should allow to catch signal CTRL_BREAK_EVENT,
            # but we to create the child process with CREATE_NEW_PROCESS_GROUP
            # to make this work, so current this is a noop for later fix
            # NOTE(sileht): is mypy posix only ?
            signal.signal(signal.SIGBREAK, self._signal_catcher)  # type: ignore[attr-defined]

    @staticmethod
    def _set_nonblock(fd: int) -> None:
        flags = fcntl.fcntl(fd, fcntl.F_GETFL, 0)
        flags |= os.O_NONBLOCK
        fcntl.fcntl(fd, fcntl.F_SETFL, flags)

    @staticmethod
    def _set_autoclose(fd: int) -> None:
        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

    def _signal_catcher(self, sig: int, frame: types.FrameType | None) -> None:
        # Needed to receive the signal via set_wakeup_fd
        pass

    def _wait_forever(self) -> None:
        # Wait forever
        while True:
            # NOTE(sileht): signals can be missed when received by non-main threads
            # (https://bugs.python.org/issue5315)
            # We use signal.set_wakeup_fd + select.select() which is more reliable
            try:
                select.select([self.signal_pipe_r], [], [])
            except OSError as e:
                if e.args[0] == errno.EINTR:
                    raise

            while True:
                try:
                    signals = os.read(self.signal_pipe_r, SIGNAL_WAKEUP_FD_READ_SIZE)
                except OSError:
                    break

                for sig in signals:
                    self._on_signal_received(sig)

                if len(signals) < SIGNAL_WAKEUP_FD_READ_SIZE:
                    break

    def _on_signal_received(self, sig: int) -> None:
        pass
